from PIL import Image
from PyQt6 import QtCore
from PyQt6.QtWidgets import QAbstractItemView, QListWidget, QListWidgetItem, QMenu

from .q_dymo_label_widgets import (
    BarcodeDymoLabelWidget,
    ImageDymoLabelWidget,
    QrDymoLabelWidget,
    TextDymoLabelWidget,
)


class QDymoLabelList(QListWidget):
    """
    A custom QListWidget for displaying and managing Dymo label widgets.
    Args:
        render_engine (RenderEngine): The render engine to use for rendering the label.
        parent (QWidget): The parent widget of this QListWidget.
    Attributes:
        renderSignal (QtCore.pyqtSignal): A signal emitted when the label is rendered.
        render_engine (RenderEngine): The render engine used for rendering the label.
    Methods:
        __init__(self, render_engine, parent=None): Initializes the QListWidget
            with the given render engine and parent.
        dropEvent(self, e) -> None: Overrides the default drop event to update
            the label rendering.
        update_render_engine(self, render_engine): Updates the render engine used
            for rendering the label.
        render_label(self): Renders the label using the current render engine and
            emits the renderSignal.
        contextMenuEvent(self, event): Overrides the default context menu event to
            add or delete label widgets.
    """

    renderSignal = QtCore.pyqtSignal(Image.Image, name="renderSignal")

    def __init__(self, render_engine, min_payload_len=0, justify="center", parent=None):
        super().__init__(parent)
        self.min_payload_len = min_payload_len
        self.justify = justify
        self.render_engine = render_engine
        self.setAlternatingRowColors(True)
        self.setDragDropMode(QAbstractItemView.DragDropMode.InternalMove)
        for item_widget in [TextDymoLabelWidget(self.render_engine)]:
            item = QListWidgetItem(self)
            item.setSizeHint(item_widget.sizeHint())
            self.addItem(item)
            self.setItemWidget(item, item_widget)
            item_widget.itemRenderSignal.connect(self.render_label)
        self.render_label()

    def dropEvent(self, e) -> None:
        """
        Overrides the default drop event to update the label rendering.
        Args:
            e (QDropEvent): The drop event.
        """
        super().dropEvent(e)
        self.render_label()

    def update_params(self, render_engine, min_payload_len=0, justify="center"):
        """
        Updates the render engine used for rendering the label.
        Args:
            justify: justification [center,left,right]
            min_payload_len: minimum payload size
            render_engine (RenderEngine): The new render engine to use.
        """
        self.min_payload_len = min_payload_len
        self.justify = justify
        self.render_engine = render_engine
        for i in range(self.count()):
            item_widget = self.itemWidget(self.item(i))
            item_widget.render_engine = render_engine
        self.render_label()

    def render_label(self):
        """
        Renders the label using the current render engine and emits the renderSignal.
        """
        bitmaps = []
        for i in range(self.count()):
            item = self.item(i)
            item_widget = self.itemWidget(self.item(i))
            if item_widget and item:
                item.setSizeHint(item_widget.sizeHint())
                bitmaps.append(item_widget.render_label())
        label_bitmap = self.render_engine.merge_render(
            bitmaps, self.min_payload_len, self.justify
        )

        self.renderSignal.emit(label_bitmap)

    def contextMenuEvent(self, event):
        """
        Overrides the default context menu event to add or delete label widgets.
        Args:
            event (QContextMenuEvent): The context menu event.
        """
        contextMenu = QMenu(self)
        add_text = contextMenu.addAction("Add Text")
        add_qr = contextMenu.addAction("Add QR")
        add_barcode = contextMenu.addAction("Add Barcode")
        add_img = contextMenu.addAction("Add Image")
        delete = contextMenu.addAction("Delete")
        menu_click = contextMenu.exec(event.globalPos())

        if menu_click == add_text:
            item = QListWidgetItem(self)
            item_widget = TextDymoLabelWidget(self.render_engine)
            item.setSizeHint(item_widget.sizeHint())
            self.addItem(item)
            self.setItemWidget(item, item_widget)
            item_widget.itemRenderSignal.connect(self.render_label)

        if menu_click == add_qr:
            item = QListWidgetItem(self)
            item_widget = QrDymoLabelWidget(self.render_engine)
            item.setSizeHint(item_widget.sizeHint())
            self.addItem(item)
            self.setItemWidget(item, item_widget)
            item_widget.itemRenderSignal.connect(self.render_label)

        if menu_click == add_barcode:
            item = QListWidgetItem(self)
            item_widget = BarcodeDymoLabelWidget(self.render_engine)
            item.setSizeHint(item_widget.sizeHint())
            self.addItem(item)
            self.setItemWidget(item, item_widget)
            item_widget.itemRenderSignal.connect(self.render_label)

        if menu_click == add_img:
            item = QListWidgetItem(self)
            item_widget = ImageDymoLabelWidget(self.render_engine)
            item.setSizeHint(item_widget.sizeHint())
            self.addItem(item)
            self.setItemWidget(item, item_widget)
            item_widget.itemRenderSignal.connect(self.render_label)
        if menu_click == delete:
            try:
                item = self.itemAt(event.pos())
                self.takeItem(self.indexFromItem(item).row())  # self.update()
            except Exception as e:
                print(f"No item selected {e}")
        self.render_label()



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

import argparse
import os

from PIL import Image, ImageOps

from . import __version__
from .constants import DEFAULT_MARGIN, USE_QR, e_qrcode
from .dymo_print_engines import DymoPrinterServer, DymoRenderEngine
from .font_config import font_filename
from .metadata import our_metadata
from .unicode_blocks import image_to_unicode
from .utils import die


def parse_args():
    # check for any text specified on the command line
    parser = argparse.ArgumentParser(description=our_metadata["Summary"])
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "text",
        nargs="+",
        help="Text Parameter, each parameter gives a new line",
        type=str,
    )
    parser.add_argument(
        "-f",
        action="count",
        help="Draw frame around the text, more arguments for thicker frame",
    )
    parser.add_argument(
        "-s",
        choices=["r", "b", "i", "n"],
        default="r",
        help="Set fonts style (regular,bold,italic,narrow)",
    )
    parser.add_argument(
        "-a",
        choices=[
            "left",
            "center",
            "right",
        ],
        default="left",
        help="Align multiline text (left,center,right)",
    )
    parser.add_argument(
        "-l", type=int, default=0, help="Specify minimum label length in mm"
    )
    parser.add_argument(
        "-j",
        choices=[
            "left",
            "center",
            "right",
        ],
        default="center",
        help=(
            "Justify content of label if minimum label length "
            "is specified (left,center,right)"
        ),
    )
    parser.add_argument("-u", nargs="?", help='Set user font, overrides "-s" parameter')
    parser.add_argument(
        "-n",
        "--preview",
        action="store_true",
        help="Unicode preview of label, do not send to printer",
    )
    parser.add_argument(
        "--preview-inverted",
        action="store_true",
        help="Unicode preview of label, colors inverted, do not send to printer",
    )
    parser.add_argument(
        "--imagemagick",
        action="store_true",
        help="Preview label with Imagemagick, do not send to printer",
    )
    parser.add_argument(
        "-qr", action="store_true", help="Printing the first text parameter as QR-code"
    )
    parser.add_argument(
        "-c",
        choices=[
            "code39",
            "code128",
            "ean",
            "ean13",
            "ean8",
            "gs1",
            "gtin",
            "isbn",
            "isbn10",
            "isbn13",
            "issn",
            "jan",
            "pzn",
            "upc",
            "upca",
        ],
        default=False,
        help="Printing the first text parameter as barcode",
    )
    parser.add_argument("-p", "--picture", help="Print the specified picture")
    parser.add_argument(
        "-m",
        type=int,
        default=DEFAULT_MARGIN,
        help=f"Override margin (default is {DEFAULT_MARGIN})",
    )
    parser.add_argument(
        "--scale", type=int, default=90, help="Scaling font factor, [0,10] [%%]"
    )
    parser.add_argument(
        "-t",
        type=int,
        choices=[6, 9, 12],
        default=12,
        help="Tape size: 6,9,12 mm, default=12mm",
    )
    return parser.parse_args()


def main():
    args = parse_args()
    print_server = DymoPrinterServer()
    render_engine = DymoRenderEngine(args.t)

    # read config file
    FONT_FILENAME = font_filename(args.s)

    labeltext = args.text

    if args.u is not None:
        if os.path.isfile(args.u):
            FONT_FILENAME = args.u
        else:
            die("Error: file '%s' not found." % args.u)

    # check if barcode, qrcode or text should be printed, use frames only on text
    if args.qr and not USE_QR:
        die("Error: %s" % e_qrcode)

    if args.c and args.qr:
        die("Error: can not print both QR and Barcode on the same label (yet)")

    bitmaps = []

    if args.qr:
        bitmaps.append(render_engine.render_qr(labeltext.pop(0)))

    elif args.c:
        bitmaps.append(render_engine.render_barcode(labeltext.pop(0), args.c))

    if labeltext:
        bitmaps.append(
            render_engine.render_text(
                labeltext, FONT_FILENAME, args.f, int(args.scale) / 100.0, args.a
            )
        )

    if args.picture:
        bitmaps.append(render_engine.render_picture(args.picture))

    margin = args.m
    justify = args.j
    min_label_mm_len: int = args.l
    min_payload_len = max(0, (min_label_mm_len * 7) - margin * 2)
    label_bitmap = render_engine.merge_render(bitmaps, min_payload_len, justify)

    # print or show the label
    if args.preview or args.preview_inverted or args.imagemagick:
        print("Demo mode: showing label..")
        # fix size, adding print borders
        label_image = Image.new(
            "L", (margin + label_bitmap.width + margin, label_bitmap.height)
        )
        label_image.paste(label_bitmap, (margin, 0))
        if args.preview or args.preview_inverted:
            label_rotated = label_bitmap.transpose(Image.ROTATE_270)
            print(image_to_unicode(label_rotated, invert=args.preview_inverted))
        if args.imagemagick:
            ImageOps.invert(label_image).show()
    else:
        print_server.print_label(label_bitmap, margin=args.m)



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===
import array
from typing import Optional

import usb

from .constants import DEFAULT_MARGIN, ESC, SYN


class DymoLabeler:
    """Create and work with a Dymo LabelManager PnP object.

    This class contains both mid-level and high-level functions. In general,
    the high-level functions should be used. However, special purpose usage
    may require the mid-level functions. That is why they are provided.
    However, they should be well understood before use. Look at the
    high-level functions for help. Each function is marked in its docstring
    with 'HLF' or 'MLF' in parentheses.

    A partial reference of the protocol is the Technical Reference for the
    LabelWriter 450:
    <https://download.dymo.com/dymo/technical-data-sheets/LW%20450%20Series%20Technical%20Reference.pdf>
    """

    @staticmethod
    def max_bytes_per_line(tape_size=12):
        return int(8 * tape_size / 12)

    # Max number of print lines to send before waiting for a response. This helps
    # to avoid timeouts due to differences between data transfer and
    # printer speeds. I added this because I kept getting "IOError: [Errno
    # 110] Connection timed out" with long labels. Using dev.default_timeout
    # (1000) and the transfer speeds available in the descriptors somewhere, a
    # sensible timeout can also be calculated dynamically.
    synwait: Optional[int]
    devout: usb.core.Endpoint
    devin: usb.core.Endpoint

    def __init__(self, devout, devin, synwait=None, tape_size=12):
        """Initialize the LabelManager object. (HLF)"""

        self.tape_size = tape_size
        self.cmd: list[int] = []
        self.response = False
        self.bytesPerLine_ = None
        self.dotTab_ = 0
        self.maxLines = 200
        self.devout = devout
        self.devin = devin
        self.synwait = synwait

    def sendCommand(self):
        """Send the already built command to the LabelManager. (MLF)"""

        if len(self.cmd) == 0:
            return

        while len(self.cmd) > 0:
            if self.synwait is None:
                cmd_to_send = self.cmd
                cmd_rest = []
            else:
                # Send a status request
                cmdBin = array.array("B", [ESC, ord("A")])
                cmdBin.tofile(self.devout)
                rspBin = self.devin.read(8)
                _ = array.array("B", rspBin).tolist()
                # Ok, we got a response. Now we can send a chunk of data

                # Compute a chunk with at most synwait SYN characters
                synCount = 0  # Number of SYN characters encountered in iteration
                pos = -1  # Index of last SYN character encountered in iteration
                while synCount < self.synwait:
                    try:
                        # Increment pos to the index of the next SYN character
                        pos += self.cmd[pos + 1 :].index(SYN) + 1
                        synCount += 1
                    except ValueError:
                        # No more SYN characters in cmd
                        pos = len(self.cmd)
                        break
                cmd_to_send = self.cmd[:pos]
                cmd_rest = self.cmd[pos:]
                print(f"Sending chunk of {len(cmd_to_send)} bytes")

            # Remove the computed chunk from the command to be processed
            self.cmd = cmd_rest

            # Send the chunk
            cmdBin = array.array("B", cmd_to_send)
            cmdBin.tofile(self.devout)

        self.cmd = []  # This looks redundant.
        if not self.response:
            return
        self.response = False
        responseBin = self.devin.read(8)
        response = array.array("B", responseBin).tolist()
        return response

    def resetCommand(self):
        """Remove a partially built command. (MLF)"""

        self.cmd = []
        self.response = False

    def buildCommand(self, cmd):
        """Add the next instruction to the command. (MLF)"""

        self.cmd += cmd

    def statusRequest(self):
        """Set instruction to get the device's status. (MLF)"""

        cmd = [ESC, ord("A")]
        self.buildCommand(cmd)
        self.response = True

    def dotTab(self, value):
        """Set the bias text height, in bytes. (MLF)"""

        if value < 0 or value > self.max_bytes_per_line(self.tape_size):
            raise ValueError
        cmd = [ESC, ord("B"), value]
        self.buildCommand(cmd)
        self.dotTab_ = value
        self.bytesPerLine_ = None

    def tapeColor(self, value):
        """Set the tape color. (MLF)"""

        if value < 0:
            raise ValueError
        cmd = [ESC, ord("C"), value]
        self.buildCommand(cmd)

    def bytesPerLine(self, value):
        """Set the number of bytes sent in the following lines. (MLF)"""

        if value < 0 or value + self.dotTab_ > self.max_bytes_per_line(self.tape_size):
            raise ValueError
        if value == self.bytesPerLine_:
            return
        cmd = [ESC, ord("D"), value]
        self.buildCommand(cmd)
        self.bytesPerLine_ = value

    def cut(self):
        """Set instruction to trigger cutting of the tape. (MLF)"""

        cmd = [ESC, ord("E")]
        self.buildCommand(cmd)

    def line(self, value):
        """Set next printed line. (MLF)"""

        self.bytesPerLine(len(value))
        cmd = [SYN] + value
        self.buildCommand(cmd)

    def chainMark(self):
        """Set Chain Mark. (MLF)"""

        self.dotTab(0)
        self.bytesPerLine(self.max_bytes_per_line(self.tape_size))
        self.line([0x99] * self.max_bytes_per_line(self.tape_size))

    def skipLines(self, value):
        """Set number of lines of white to print. (MLF)"""

        if value <= 0:
            raise ValueError
        self.bytesPerLine(0)
        cmd = [SYN] * value
        self.buildCommand(cmd)

    def initLabel(self):
        """Set the label initialization sequence. (MLF)"""

        cmd = [0x00] * 8
        self.buildCommand(cmd)

    def getStatus(self):
        """Ask for and return the device's status. (HLF)"""

        self.statusRequest()
        response = self.sendCommand()
        print(response)

    def printLabel(self, lines, margin=DEFAULT_MARGIN):
        """Print the label described by lines. (Automatically split label if
        larger than maxLines)"""

        while len(lines) > self.maxLines + 1:
            self.rawPrintLabel(lines[0 : self.maxLines], margin=0)
            del lines[0 : self.maxLines]
        self.rawPrintLabel(lines, margin=margin)

    def rawPrintLabel(self, lines, margin=DEFAULT_MARGIN):
        """Print the label described by lines. (HLF)"""

        # optimize the matrix for the dymo label printer
        dottab = 0
        while [] not in lines and max(line[0] for line in lines) == 0:
            lines = [line[1:] for line in lines]
            dottab += 1
        for line in lines:
            while len(line) > 0 and line[-1] == 0:
                del line[-1]

        self.initLabel
        self.tapeColor(0)
        self.dotTab(dottab)
        for line in lines:
            self.line(line)
        if margin > 0:
            self.skipLines(margin * 2)
        self.statusRequest()
        response = self.sendCommand()
        print(f"Post-send response: {response}")



import os

from PyQt6 import QtCore
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import (
    QComboBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QPlainTextEdit,
    QPushButton,
    QSpinBox,
    QWidget,
)

import dymoprint_fonts

from .font_config import parse_fonts


class BaseDymoLabelWidget(QWidget):
    """
    A base class for creating Dymo label widgets.
    Signals:
    --------
    itemRenderSignal : PyQtSignal
        Signal emitted when the content of the label is changed.
    Methods:
    --------
    content_changed()
        Emits the itemRenderSignal when the content of the label is changed.
    render_label()
        Abstract method to be implemented by subclasses for rendering the label.
    """

    itemRenderSignal = QtCore.pyqtSignal(name="itemRenderSignal")

    def content_changed(self):
        """
        Emits the itemRenderSignal when the content of the label is changed.
        """
        self.itemRenderSignal.emit()

    def render_label(self):
        """
        Abstract method to be implemented by subclasses for rendering the label.
        """
        pass


class TextDymoLabelWidget(BaseDymoLabelWidget):
    """
    A widget for rendering text on a Dymo label.
    Args:
        render_engine (RenderEngine): The rendering engine to use.
        parent (QWidget): The parent widget of this widget.
    Attributes:
        render_engine (RenderEngine): The rendering engine used by this widget.
        label (QPlainTextEdit): The text label to be rendered on the Dymo label.
        font_style (QComboBox): The font style selection dropdown.
        font_size (QSpinBox): The font size selection spinner.
        draw_frame (QSpinBox): The frame width selection spinner.
    Signals:
        itemRenderSignal: A signal emitted when the content of the label changes.
    """

    def __init__(self, render_engine, parent=None):
        super().__init__(parent)
        self.render_engine = render_engine

        self.label = QPlainTextEdit("text")
        self.label.setFixedHeight(15 * (len(self.label.toPlainText().splitlines()) + 2))
        self.setFixedHeight(self.label.height() + 10)
        self.font_style = QComboBox()
        self.font_size = QSpinBox()
        self.font_size.setMaximum(150)
        self.font_size.setMinimum(0)
        self.font_size.setSingleStep(1)
        self.font_size.setValue(90)
        self.draw_frame = QSpinBox()
        self.align = QComboBox()

        self.align.addItems(["left", "center", "right"])

        for name, font_path in parse_fonts():
            self.font_style.addItem(name, font_path)
            if "Regular" in name:
                self.font_style.setCurrentText(name)

        layout = QHBoxLayout()
        ICON_DIR = os.path.dirname(dymoprint_fonts.__file__)
        item_icon = QLabel()
        item_icon.setPixmap(
            QIcon(os.path.join(ICON_DIR, "txt_icon.png")).pixmap(32, 32)
        )
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        layout.addWidget(QLabel("Font:"))
        layout.addWidget(self.font_style)
        layout.addWidget(QLabel("Size [%]:"))
        layout.addWidget(self.font_size)
        layout.addWidget(QLabel("Frame Width:"))
        layout.addWidget(self.draw_frame)
        layout.addWidget(QLabel("Alignment:"))
        layout.addWidget(self.align)
        self.label.textChanged.connect(self.content_changed)
        self.draw_frame.valueChanged.connect(self.content_changed)
        self.font_size.valueChanged.connect(self.content_changed)
        self.font_style.currentTextChanged.connect(self.content_changed)
        self.align.currentTextChanged.connect(self.content_changed)
        self.setLayout(layout)

    def content_changed(self):
        """
        Updates the height of the label and emits the itemRenderSignal when the
        content of the label changes.
        """
        self.label.setFixedHeight(15 * (len(self.label.toPlainText().splitlines()) + 2))
        self.setFixedHeight(self.label.height() + 10)
        self.itemRenderSignal.emit()

    def render_label(self):
        """
        Renders the label using the current settings.
        Returns:
            QImage: The rendered label image.
        Raises:
            QMessageBox.warning: If the rendering fails.
        """
        try:
            render = self.render_engine.render_text(
                labeltext=self.label.toPlainText().splitlines(),
                font_file_name=self.font_style.currentData(),
                frame_width=self.draw_frame.value(),
                font_size_ratio=self.font_size.value() / 100.0,
                align=self.align.currentText(),
            )
            return render
        except BaseException as err:
            QMessageBox.warning(self, "TextDymoLabelWidget render fail!", f"{err}")
            return self.render_engine.render_empty()


class QrDymoLabelWidget(BaseDymoLabelWidget):
    """
    A widget for rendering QR codes on Dymo labels.
    Args:
        render_engine (RenderEngine): The render engine to use for rendering
            the QR code.
        parent (QWidget, optional): The parent widget. Defaults to None.
    """

    def __init__(self, render_engine, parent=None):
        """
        Initializes the QrDymoLabelWidget.
        Args:
            render_engine (RenderEngine): The render engine to use for rendering
                the QR code.
            parent (QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.render_engine = render_engine

        self.label = QLineEdit("")
        layout = QHBoxLayout()
        ICON_DIR = os.path.dirname(dymoprint_fonts.__file__)
        item_icon = QLabel()
        item_icon.setPixmap(QIcon(os.path.join(ICON_DIR, "qr_icon.png")).pixmap(32, 32))
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        self.label.textChanged.connect(self.content_changed)
        self.setLayout(layout)

    def render_label(self):
        """
        Renders the QR code on the Dymo label.
        Returns:
            bytes: The rendered QR code as bytes.
        Raises:
            QMessageBox.warning: If the rendering fails.
        """
        try:
            render = self.render_engine.render_qr(self.label.text())
            return render

        except BaseException as err:
            QMessageBox.warning(self, "QrDymoLabelWidget render fail!", f"{err}")
            return self.render_engine.render_empty()


class BarcodeDymoLabelWidget(BaseDymoLabelWidget):
    """
    A widget for rendering barcode labels using the Dymo label printer.
    Args:
        render_engine (DymoRenderEngine): An instance of the DymoRenderEngine class.
        parent (QWidget): The parent widget of this widget.
    Attributes:
        render_engine (DymoRenderEngine): An instance of the DymoRenderEngine class.
        label (QLineEdit): A QLineEdit widget for entering the content of the
            barcode label.
        codding (QComboBox): A QComboBox widget for selecting the type of barcode
            to render.
    Signals:
        content_changed(): Emitted when the content of the label or the selected
            barcode type changes.
    Methods:
        __init__(self, render_engine, parent=None): Initializes the widget.
        render_label(self): Renders the barcode label using the current content
            and barcode type.
    """

    def __init__(self, render_engine, parent=None):
        super().__init__(parent)
        self.render_engine = render_engine

        self.label = QLineEdit("")
        layout = QHBoxLayout()
        ICON_DIR = os.path.dirname(dymoprint_fonts.__file__)
        item_icon = QLabel()
        item_icon.setPixmap(
            QIcon(os.path.join(ICON_DIR, "barcode_icon.png")).pixmap(32, 32)
        )
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)
        self.codding = QComboBox()
        self.codding.addItems(
            [
                "code39",
                "code128",
                "ean",
                "ean13",
                "ean8",
                "gs1",
                "gtin",
                "isbn",
                "isbn10",
                "isbn13",
                "issn",
                "jan",
                "pzn",
                "upc",
                "upca",
            ]
        )

        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        layout.addWidget(QLabel("Codding:"))
        layout.addWidget(self.codding)

        self.label.textChanged.connect(self.content_changed)
        self.codding.currentTextChanged.connect(self.content_changed)
        self.setLayout(layout)

    def render_label(self):
        """
        Renders the barcode label using the current content and barcode type.
        Returns:
            QPixmap: A QPixmap object representing the rendered barcode label.
        """
        try:
            render = self.render_engine.render_barcode(
                self.label.text(), self.codding.currentText()
            )
            return render

        except BaseException as err:
            QMessageBox.warning(self, "BarcodeDymoLabelWidget render fail!", f"{err}")
            return self.render_engine.render_empty()


class ImageDymoLabelWidget(BaseDymoLabelWidget):
    """
    A widget for rendering image-based Dymo labels.
    Args:
        render_engine (RenderEngine): The render engine to use for rendering the label.
        parent (QWidget, optional): The parent widget. Defaults to None.
    """

    def __init__(self, render_engine, parent=None):
        """
        Initializes the ImageDymoLabelWidget.
        Args:
            render_engine (RenderEngine): The render engine to use for rendering
                the label.
            parent (QWidget, optional): The parent widget. Defaults to None.
        """
        super().__init__(parent)
        self.render_engine = render_engine

        self.label = QLineEdit("")
        layout = QHBoxLayout()
        ICON_DIR = os.path.dirname(dymoprint_fonts.__file__)
        item_icon = QLabel()
        item_icon.setPixmap(
            QIcon(os.path.join(ICON_DIR, "img_icon.png")).pixmap(32, 32)
        )
        item_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignHCenter)

        button = QPushButton("Select file")
        file_dialog = QFileDialog()
        button.clicked.connect(
            lambda: self.label.setText(
                os.path.abspath(file_dialog.getOpenFileName()[0])
            )
        )

        layout.addWidget(item_icon)
        layout.addWidget(self.label)
        layout.addWidget(button)

        self.label.textChanged.connect(self.content_changed)
        self.setLayout(layout)

    def render_label(self):
        """
        Renders the label using the render engine and the selected image file.
        Returns:
            QPixmap: The rendered label as a QPixmap.
        """
        try:
            render = self.render_engine.render_picture(self.label.text())
            return render
        except BaseException as err:
            QMessageBox.warning(self, "ImageDymoLabelWidget render fail!", f"{err}")
            return self.render_engine.render_empty()



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

# On systems with access to sysfs under /sys, this script will use the three
# variables DEV_CLASS, DEV_VENDOR, and DEV_PRODUCT to find the device file
# under /dev automatically. This behavior can be overridden by setting the
# variable DEV_NODE to the device file path. This is intended for cases, where
# either sysfs is unavailable or unusable by this script for some reason.
# Please beware that DEV_NODE must be set to None when not used, else you will
# be bitten by the NameError exception.

try:
    from pyqrcode import QRCode

    USE_QR = True
    e_qrcode = None
except ImportError as error:
    e_qrcode = error
    USE_QR = False
    QRCode = None


DESCRIPTION = (
    "Linux Software to print with LabelManager PnP from Dymo\nwritten in Python"
)
UNCONFIRMED_MESSAGE = (
    "WARNING: This device is not confirmed to work with this software. Please "
    "report your experiences in https://github.com/computerlyrik/dymoprint/issues/44"
)
SUPPORTED_PRODUCTS = {
    0x1001: "LabelManager PnP (no mode switch)",
    0x1002: "LabelManager PnP (mode switch)",
    0x1003: f"LabelManager 420P (no mode switch) {UNCONFIRMED_MESSAGE}",
    0x1004: f"LabelManager 420P (mode switch) {UNCONFIRMED_MESSAGE}",
    0x1005: "LabelManager 280 (no mode switch)",
    0x1006: "LabelManager 280 (no mode switch)",
    0x1007: f"LabelManager Wireless PnP (no mode switch) {UNCONFIRMED_MESSAGE}",
    0x1008: f"LabelManager Wireless PnP (mode switch) {UNCONFIRMED_MESSAGE}",
    0x1009: f"MobileLabeler {UNCONFIRMED_MESSAGE}",
}
DEV_VENDOR = 0x0922

PRINTER_INTERFACE_CLASS = 0x07
HID_INTERFACE_CLASS = 0x03

# Escape character preceeding all commands
ESC = 0x1B

# Synchronization character preceding uncompressed print data
SYN = 0x16

FONT_SIZERATIO = 7 / 8

DEFAULT_FONT_STYLE = "regular"

DEFAULT_MARGIN = 56

FLAG_TO_STYLE = {
    "r": "regular",
    "b": "bold",
    "i": "italic",
    "n": "narrow",
}



import sys

if sys.version_info[:2] >= (3, 8):
    from importlib.metadata import metadata as md
else:
    from importlib_metadata import metadata as md

dist_name = __name__.split(".")[0]
our_metadata = md(dist_name)
__version__ = our_metadata["Version"]



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

from typing import Optional

from barcode.writer import BaseWriter
from PIL import Image, ImageDraw


def mm2px(mm, dpi=25.4):
    return (mm * dpi) / 25.4


class BarcodeImageWriter(BaseWriter):
    _draw: Optional[ImageDraw.ImageDraw]

    def __init__(self):
        super().__init__(self._init, self._paint_module, None, self._finish)
        self.format = "PNG"
        self.dpi = 25.4
        self._image = None
        self._draw = None
        self.vertical_margin = 0

    def calculate_size(self, modules_per_line, number_of_lines, dpi=25.4):
        width = 2 * self.quiet_zone + modules_per_line * self.module_width
        height = self.vertical_margin * 2 + self.module_height * number_of_lines
        return int(mm2px(width, dpi)), int(mm2px(height, dpi))

    def render(self, code):
        """Renders the barcode to whatever the inheriting writer provides,
        using the registered callbacks.

        :parameters:
            code : List
                List of strings matching the writer spec
                (only contain 0 or 1).
        """
        if self._callbacks["initialize"] is not None:
            self._callbacks["initialize"](code)
        ypos = self.vertical_margin
        for cc, line in enumerate(code):
            """
            Pack line to list give better gfx result, otherwise in can
            result in aliasing gaps
            '11010111' -> [2, -1, 1, -1, 3]
            """
            line += " "
            c = 1
            mlist = []
            for i in range(0, len(line) - 1):
                if line[i] == line[i + 1]:
                    c += 1
                else:
                    if line[i] == "1":
                        mlist.append(c)
                    else:
                        mlist.append(-c)
                    c = 1
            # Left quiet zone is x startposition
            xpos = self.quiet_zone
            for mod in mlist:
                if mod < 1:
                    color = self.background
                else:
                    color = self.foreground
                # remove painting for background colored tiles?
                self._callbacks["paint_module"](
                    xpos, ypos, self.module_width * abs(mod), color
                )
                xpos += self.module_width * abs(mod)
            # Add right quiet zone to every line, except last line,
            # quiet zone already provided with background,
            # should it be removed complety?
            if (cc + 1) != len(code):
                self._callbacks["paint_module"](
                    xpos, ypos, self.quiet_zone, self.background
                )
            ypos += self.module_height
        return self._callbacks["finish"]()

    def _init(self, code):
        size = self.calculate_size(len(code[0]), len(code), self.dpi)
        self._image = Image.new("1", size, self.background)
        self._draw = ImageDraw.Draw(self._image)

    def _paint_module(self, xpos, ypos, width, color):
        size = [
            (mm2px(xpos, self.dpi), mm2px(ypos, self.dpi)),
            (
                mm2px(xpos + width, self.dpi),
                mm2px(ypos + self.module_height, self.dpi),
            ),
        ]
        self._draw.rectangle(size, outline=color, fill=color)

    def _finish(self):
        return self._image

    def save(self, filename, output):
        filename = f"{filename}.{self.format.lower()}"
        output.save(filename, self.format.upper())
        return filename



import os
import sys
import traceback

from PIL import Image, ImageOps, ImageQt
from PyQt6 import QtCore
from PyQt6.QtCore import QSize, Qt
from PyQt6.QtGui import QColor, QIcon, QPainter, QPixmap
from PyQt6.QtWidgets import (
    QApplication,
    QComboBox,
    QGraphicsDropShadowEffect,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QSpinBox,
    QToolBar,
    QVBoxLayout,
    QWidget,
)
from usb.core import USBError

import dymoprint_fonts

from .constants import DEFAULT_MARGIN
from .dymo_print_engines import DymoPrinterServer, DymoRenderEngine
from .q_dymo_labels_list import QDymoLabelList


class DymoPrintWindow(QWidget):
    def __init__(self):
        super().__init__()
        self.print_server = DymoPrinterServer()
        self.render_engine = DymoRenderEngine(12)
        self.label_bitmap = Image

        self.window_layout = QVBoxLayout()
        self.list = QDymoLabelList(self.render_engine)
        self.label_render = QLabel()
        self.print_button = QPushButton()
        self.margin = QSpinBox()
        self.tape_size = QComboBox()
        self.foreground_color = QComboBox()
        self.background_color = QComboBox()
        self.min_label_len = QSpinBox()
        self.justify = QComboBox()

        self.init_elements()
        self.init_connections()
        self.init_layout()

        self.list.render_label()

    def init_elements(self):
        self.setWindowTitle("DymoPrint GUI")
        ICON_DIR = os.path.dirname(dymoprint_fonts.__file__)
        self.setWindowIcon(QIcon(os.path.join(ICON_DIR, "gui_icon.png")))
        self.setGeometry(200, 200, 1100, 400)
        printer_icon = QIcon.fromTheme("printer")
        self.print_button.setIcon(printer_icon)
        self.print_button.setFixedSize(64, 64)
        self.print_button.setIconSize(QSize(48, 48))

        shadow = QGraphicsDropShadowEffect()
        shadow.setBlurRadius(15)
        self.label_render.setGraphicsEffect(shadow)

        self.margin.setMinimum(20)
        self.margin.setMaximum(1000)
        self.margin.setValue(DEFAULT_MARGIN)
        self.tape_size.addItem("12", 12)
        self.tape_size.addItem("9", 9)
        self.tape_size.addItem("6", 6)
        self.min_label_len.setMinimum(0)
        self.min_label_len.setMaximum(1000)
        self.justify.addItems(["center", "left", "right"])

        self.foreground_color.addItems(
            ["black", "white", "yellow", "blue", "red", "green"]
        )
        self.background_color.addItems(
            ["white", "black", "yellow", "blue", "red", "green"]
        )

    def init_connections(self):
        self.margin.valueChanged.connect(self.list.render_label)
        self.tape_size.currentTextChanged.connect(self.update_params)
        self.min_label_len.valueChanged.connect(self.update_params)
        self.justify.currentTextChanged.connect(self.update_params)
        self.foreground_color.currentTextChanged.connect(self.list.render_label)
        self.background_color.currentTextChanged.connect(self.list.render_label)
        self.list.renderSignal.connect(self.update_label_render)
        self.print_button.clicked.connect(self.print_label)

    def init_layout(self):
        settings_widget = QToolBar(self)
        settings_widget.addWidget(QLabel("Margin:"))
        settings_widget.addWidget(self.margin)
        settings_widget.addSeparator()
        settings_widget.addWidget(QLabel("Tape Size:"))
        settings_widget.addWidget(self.tape_size)
        settings_widget.addSeparator()
        settings_widget.addWidget(QLabel("Min Label Len [mm]:"))
        settings_widget.addWidget(self.min_label_len)
        settings_widget.addSeparator()
        settings_widget.addWidget(QLabel("Justify:"))
        settings_widget.addWidget(self.justify)
        settings_widget.addSeparator()
        settings_widget.addWidget(QLabel("Tape Colors: "))
        settings_widget.addWidget(self.foreground_color)
        settings_widget.addWidget(QLabel(" on "))
        settings_widget.addWidget(self.background_color)

        render_widget = QWidget(self)
        render_layout = QHBoxLayout(render_widget)
        render_layout.addWidget(self.label_render)
        render_layout.addWidget(self.print_button)
        render_layout.setAlignment(
            self.label_render, QtCore.Qt.AlignmentFlag.AlignCenter
        )

        self.window_layout.addWidget(settings_widget)
        self.window_layout.addWidget(self.list)
        self.window_layout.addWidget(render_widget)
        self.setLayout(self.window_layout)

    def update_params(self):
        self.render_engine = DymoRenderEngine(self.tape_size.currentData())
        justify = self.justify.currentText()
        min_label_mm_len: int = self.min_label_len.value()
        min_payload_len = max(0, (min_label_mm_len * 7) - self.margin.value() * 2)
        self.list.update_params(self.render_engine, min_payload_len, justify)

    def update_label_render(self, label_bitmap):
        self.label_bitmap = label_bitmap
        label_image = Image.new(
            "L",
            (
                self.margin.value() + self.label_bitmap.width + self.margin.value(),
                self.label_bitmap.height,
            ),
        )
        label_image.paste(self.label_bitmap, (self.margin.value(), 0))
        label_image_inv = ImageOps.invert(label_image).copy()
        qim = ImageQt.ImageQt(label_image_inv)
        q_image = QPixmap.fromImage(qim)

        mask = q_image.createMaskFromColor(
            QColor("255, 255, 255"), Qt.MaskMode.MaskOutColor
        )
        q_image.fill(QColor(self.background_color.currentText()))
        p = QPainter(q_image)
        p.setPen(QColor(self.foreground_color.currentText()))
        p.drawPixmap(q_image.rect(), mask, mask.rect())
        p.end()

        self.label_render.setPixmap(q_image)
        self.label_render.adjustSize()

    def print_label(self):
        try:
            self.print_server.print_label(self.label_bitmap, self.margin.value())
        except (RuntimeError, USBError) as err:
            print(traceback.format_exc())
            QMessageBox.warning(
                self, "Printing Failed!", f"{err}\n\n{traceback.format_exc()}"
            )


def main():
    app = QApplication(sys.argv)
    window = DymoPrintWindow()
    window.show()
    sys.exit(app.exec())



# === LICENSE STATEMENT ===
# Copyright (c) 2011 Sebastian J. Bronner <waschtl@sbronner.com>
#
# Copying and distribution of this file, with or without modification, are
# permitted in any medium without royalty provided the copyright notice and
# this notice are preserved.
# === END LICENSE STATEMENT ===

import contextlib
import sys
from typing import NoReturn

from PIL import ImageDraw


def die(message=None) -> NoReturn:
    if message:
        print(message, file=sys.stderr)
        raise RuntimeError(message)
    sys.exit(1)


""" scaling pixel up, input: (x,y),scale-factor """


def scaling(pix, sc):
    return [(pix[0] + i, pix[1] + j) for i in range(sc) for j in range(sc)]


@contextlib.contextmanager
def draw_image(bitmap):
    drawobj = ImageDraw.Draw(bitmap)
    assert isinstance(drawobj, ImageDraw.ImageDraw)
    try:
        yield drawobj
    finally:
        del drawobj



import os
import re
from configparser import ConfigParser

from appdirs import user_config_dir

import dymoprint_fonts

from .constants import DEFAULT_FONT_STYLE, FLAG_TO_STYLE
from .utils import die


def font_filename(flag):
    # The directory of the dymoprints_fonts package
    DEFAULT_FONT_DIR = os.path.dirname(dymoprint_fonts.__file__)

    # Default values
    style_to_file = {
        "regular": os.path.join(DEFAULT_FONT_DIR, "Carlito-Regular.ttf"),
        "bold": os.path.join(DEFAULT_FONT_DIR, "Carlito-Bold.ttf"),
        "italic": os.path.join(DEFAULT_FONT_DIR, "Carlito-Italic.ttf"),
        "narrow": os.path.join(DEFAULT_FONT_DIR, "Carlito-BoldItalic.ttf"),
    }

    conf = ConfigParser(style_to_file)
    CONFIG_FILE = os.path.join(user_config_dir(), "dymoprint.ini")
    if conf.read(CONFIG_FILE):
        # reading FONTS section
        if "FONTS" not in conf.sections():
            die('! config file "%s" not valid. Please change or remove.' % CONFIG_FILE)
        for style in style_to_file.keys():
            style_to_file[style] = conf.get("FONTS", style)

    return style_to_file[FLAG_TO_STYLE.get(flag, DEFAULT_FONT_STYLE)]


def parse_fonts() -> dict:
    DEFAULT_FONT_DIR = os.path.dirname(dymoprint_fonts.__file__)
    fonts = list()
    for f in os.listdir(DEFAULT_FONT_DIR):
        m = re.match(r"(.*-.*).ttf", f)
        if m:
            fonts.append((m.group(1), os.path.join(DEFAULT_FONT_DIR, f)))
    return fonts



from __future__ import annotations

import array
import math
import os
import platform
from typing import NoReturn

import barcode as barcode_module
import usb
from PIL import Image, ImageFont, ImageOps

from . import DymoLabeler
from .barcode_writer import BarcodeImageWriter
from .constants import (
    DEFAULT_MARGIN,
    DEV_VENDOR,
    HID_INTERFACE_CLASS,
    PRINTER_INTERFACE_CLASS,
    SUPPORTED_PRODUCTS,
    UNCONFIRMED_MESSAGE,
    QRCode,
)
from .utils import die, draw_image, scaling

GITHUB_LINK = "<https://github.com/computerlyrik/dymoprint/pull/56>"


def device_info(dev: usb.core.Device) -> str:
    try:
        dev.manufacturer
    except ValueError:
        instruct_on_access_denied(dev)
    res = ""
    res += f"{repr(dev)}\n"
    res += f"  manufacturer: {dev.manufacturer}\n"
    res += f"  product: {dev.product}\n"
    res += f"  serial: {dev.serial_number}\n"
    configs = dev.configurations()
    if configs:
        res += "  configurations:\n"
        for cfg in configs:
            res += f"  - {repr(cfg)}\n"
            intfs = cfg.interfaces()
            if intfs:
                res += "    interfaces:\n"
                for intf in intfs:
                    res += f"    - {repr(intf)}\n"
    return res


def instruct_on_access_denied(dev: usb.core.Device) -> NoReturn:
    system = platform.system()
    if system == "Linux":
        instruct_on_access_denied_linux(dev)
    elif system == "Windows":
        raise RuntimeError(
            "Couldn't access the device. Please make sure that the "
            "device driver is set to WinUSB. This can be accomplished "
            "with Zadig <https://zadig.akeo.ie/>."
        )
    elif system == "Darwin":
        raise RuntimeError(
            f"Could not access {dev}. Thanks for bravely trying this on a Mac. You "
            f"are in uncharted territory. It would be appreciated if you share the "
            f"results of your experimentation at {GITHUB_LINK}."
        )
    else:
        raise RuntimeError(f"Unknown platform {system}")


def instruct_on_access_denied_linux(dev: usb.core.Device) -> NoReturn:
    # try:
    #     os_release = platform.freedesktop_os_release()
    # except OSError:
    #     os_release = {}
    # dists_with_empties = [os_release.get("ID", "")] + os_release.get(
    #     "ID_LIKE", ""
    # ).split(" ")
    # dists = [dist for dist in dists_with_empties if dist]
    # if "arch" in dists:
    #     restart_udev_command = "sudo udevadm control --reload"
    # elif "ubuntu" in dists or "debian" in dists:
    #     restart_udev_command = "sudo systemctl restart udev.service"
    # # detect whether we are in arch linux or ubuntu linux
    # if Path("/etc/arch-release").exists():
    #     restart_udev_command = "sudo udevadm control --reload"
    # elif Path("/etc/lsb-release").exists():
    #     restart_udev_command = "sudo systemctl restart udev.service"
    # else:
    #     restart_udev_command = None

    lines = []
    lines.append(
        "You do not have sufficient access to the "
        "device. You probably want to add the a udev rule in "
        "/etc/udev/rules.d with the following command:"
    )
    lines.append("")
    udev_rule = ", ".join(
        [
            'ACTION=="add"',
            'SUBSYSTEMS=="usb"',
            f'ATTRS{{idVendor}}=="{dev.idVendor:04x}"',
            f'ATTRS{{idProduct}}=="{dev.idProduct:04x}"',
            'MODE="0660"',
            'GROUP="plugdev"',
            'TAG+="uaccess"',
        ]
    )
    lines.append(
        f"  echo '{udev_rule}' "
        f"| sudo tee /etc/udev/rules.d/91-dymo-{dev.idProduct:x}.rules"
    )
    lines.append("")
    lines.append("Next refresh udev with:")
    lines.append("")
    lines.append("  sudo udevadm control --reload-rules")
    lines.append('  sudo udevadm trigger --attr-match=idVendor="0922"')
    lines.append("")
    lines.append(
        "Finally, turn your device off and back "
        "on again to activate the new permissions."
    )
    lines.append("")
    lines.append(
        f"If this still does not resolve the problem, you might need to reboot. "
        f"In case rebooting is necessary, please report this at {GITHUB_LINK}. "
        f"We are still trying to figure out a simple procedure which works "
        f"for everyone. In case you still cannot connect, "
        f"or if you have any information or ideas, please post them at "
        f"that link."
    )
    raise RuntimeError("\n\n" + "\n".join(lines) + "\n")


class DymoRenderEngine:
    def __init__(self, tape_size=12):
        """
        Initializes a DymoRenderEngine object with a specified tape size.

        Args:
            tape_size (int): The size of the tape in millimeters. Default is 12mm.
        """
        self.tape_size = tape_size

    def render_empty(self, label_len=1):
        """
        Renders an empty label image.

        Returns:
            Image: An empty label image.
        """
        label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
        return Image.new("1", (label_len, label_height))

    def render_qr(self, qr_input_text):
        """
        Renders a QR code image from the input text.

        Args:
            qr_input_text (str): The input text to be encoded in the QR code.

        Returns:
            Image: A QR code image.
        """
        label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
        if len(qr_input_text) == 0:
            return Image.new("1", (1, label_height))

        # create QR object from first string
        code = QRCode(qr_input_text, error="M")
        qr_text = code.text(quiet_zone=1).split()

        # create an empty label image
        qr_scale = label_height // len(qr_text)
        qr_offset = (label_height - len(qr_text) * qr_scale) // 2
        label_width = len(qr_text) * qr_scale

        if not qr_scale:
            die(
                "Error: too much information to store in the QR code, points "
                "are smaller than the device resolution"
            )

        code_bitmap = Image.new("1", (label_width, label_height))

        with draw_image(code_bitmap) as label_draw:
            # write the qr-code into the empty image
            for i, line in enumerate(qr_text):
                for j in range(len(line)):
                    if line[j] == "1":
                        pix = scaling(
                            (j * qr_scale, i * qr_scale + qr_offset), qr_scale
                        )
                        label_draw.point(pix, 255)
        return code_bitmap

    def render_barcode(self, barcode_input_text, bar_code_type):
        """
        Renders a barcode image from the input text and barcode type.

        Args:
            barcode_input_text (str): The input text to be encoded in the barcode.
            bar_code_type (str): The type of barcode to be rendered.

        Returns:
            Image: A barcode image.
        """
        label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
        if len(barcode_input_text) == 0:
            return Image.new("1", (1, label_height))

        code = barcode_module.get(
            bar_code_type, barcode_input_text, writer=BarcodeImageWriter()
        )
        code_bitmap = code.render(
            {
                "font_size": 0,
                "vertical_margin": 8,
                "module_height": (DymoLabeler.max_bytes_per_line(self.tape_size) * 8)
                - 16,
                "module_width": 2,
                "background": "black",
                "foreground": "white",
            }
        )
        return code_bitmap

    def render_text(
        self,
        labeltext: list[str],
        font_file_name: str,
        frame_width,
        font_size_ratio=0.9,
        align="left",
    ):
        """
        Renders a text image from the input text, font file name, frame width, and
        font size ratio.

        Args:
            labeltext (list[str]): The input text to be rendered.
            font_file_name (str): The name of the font file to be used.
            frame_width (int): The width of the frame around the text.
            font_size_ratio (float): The ratio of font size to line height. Default
                is 1.

        Returns:
            Image: A text image.
        """
        if type(labeltext) is str:
            labeltext = [labeltext]

        if len(labeltext) == 0:
            labeltext = [" "]

        # create an empty label image
        label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
        line_height = float(label_height) / len(labeltext)
        fontsize = int(round(line_height * font_size_ratio))

        font_offset = int((line_height - fontsize) / 2)

        if frame_width:
            frame_width = min(frame_width, font_offset)
            frame_width = min(frame_width, 3)

        font = ImageFont.truetype(font_file_name, fontsize)
        label_width = max(font.getsize(line)[0] for line in labeltext) + (
            font_offset * 2
        )
        text_bitmap = Image.new("1", (label_width, label_height))
        with draw_image(text_bitmap) as label_draw:
            # draw frame into empty image
            if frame_width:
                label_draw.rectangle(
                    ((0, 4), (label_width - 1, label_height - 4)), fill=255
                )
                label_draw.rectangle(
                    (
                        (frame_width, 4 + frame_width),
                        (
                            label_width - (frame_width + 1),
                            label_height - (frame_width + 4),
                        ),
                    ),
                    fill=0,
                )

            # write the text into the empty image
            multiline_text = "\n".join(labeltext)
            label_draw.multiline_text(
                (label_width / 2, label_height / 2),
                multiline_text,
                align=align,
                anchor="mm",
                font=font,
                fill=255,
            )
        return text_bitmap

    def render_picture(self, picture_path: str):
        """
        Renders a picture image from the input picture path.

        Args:
            picture_path (str): The path of the picture to be rendered.

        Returns:
            Image: A picture image.
        """
        if len(picture_path):
            if os.path.exists(picture_path):
                label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
                with Image.open(picture_path) as img:
                    if img.height > label_height:
                        ratio = label_height / img.height
                        img = img.resize(
                            (int(math.ceil(img.width * ratio)), label_height)
                        )

                    img = img.convert("L", palette=Image.AFFINE)
                    return ImageOps.invert(img).convert("1")
            else:
                die(f"picture path:{picture_path}  doesn't exist ")
        label_height = DymoLabeler.max_bytes_per_line(self.tape_size) * 8
        return Image.new("1", (1, label_height))

    def merge_render(self, bitmaps, min_payload_len=0, justify="center"):
        """
        Merges multiple images into a single image.

        Args:
            bitmaps (list[Image]): A list of images to be merged.

        Returns:
            Image: A merged image.
        """
        if len(bitmaps) > 1:
            padding = 4
            label_bitmap = Image.new(
                "1",
                (
                    sum(b.width for b in bitmaps) + padding * (len(bitmaps) - 1),
                    bitmaps[0].height,
                ),
            )
            offset = 0
            for bitmap in bitmaps:
                label_bitmap.paste(bitmap, box=(offset, 0))
                offset += bitmap.width + padding
        elif len(bitmaps) == 0:
            label_bitmap = self.render_empty(max(min_payload_len, 1))
        else:
            label_bitmap = bitmaps[0]

        if min_payload_len > label_bitmap.width:
            offset = 0
            if justify == "center":
                offset = max(0, int((min_payload_len - label_bitmap.width) / 2))
            if justify == "right":
                offset = max(0, int(min_payload_len - label_bitmap.width))
            out_label_bitmap = Image.new(
                "1",
                (
                    min_payload_len,
                    label_bitmap.height,
                ),
            )
            out_label_bitmap.paste(label_bitmap, box=(offset, 0))
            return out_label_bitmap

        return label_bitmap


class DymoPrinterServer:
    @staticmethod
    def print_label(label_bitmap, margin=DEFAULT_MARGIN, tape_size: int = 12):
        """
        Prints a label using a Dymo labeler object.

        :param label_bitmap: The image to be printed as a label.
        :type label_bitmap: Image
        :param margin: The margin size in dots.
        :type margin: int, optional
        :param tape_size: The size of the tape in millimeters. Default is 12.
        :type tape_size: int, optional
        :return: None
        :rtype: None
        """
        # convert the image to the proper matrix for the dymo labeler object
        label_rotated = label_bitmap.transpose(Image.ROTATE_270)
        labelstream = label_rotated.tobytes()
        label_stream_row_length = int(math.ceil(label_bitmap.height / 8))
        if len(labelstream) // label_stream_row_length != label_bitmap.width:
            die(
                "An internal problem was encountered while processing the label "
                "bitmap!"
            )
        label_rows = [
            labelstream[i : i + label_stream_row_length]
            for i in range(0, len(labelstream), label_stream_row_length)
        ]
        label_matrix = [
            array.array("B", label_row).tolist() for label_row in label_rows
        ]

        syn_wait = 64

        dymo_devs = list(usb.core.find(idVendor=DEV_VENDOR, find_all=True))
        if len(dymo_devs) == 0:
            print(f"No Dymo devices found (expected vendor {hex(DEV_VENDOR)})")
            for dev in usb.core.find(find_all=True):
                print(
                    f"- Vendor ID: {hex(dev.idVendor):6}  "
                    f"Product ID: {hex(dev.idProduct)}"
                )
            die("Unable to open device.")
        if len(dymo_devs) > 1:
            print("Found multiple Dymo devices:")
            for dev in dymo_devs:
                print(device_info(dev))
            print("Using first device.")
            dev = dymo_devs[0]
        else:
            dev = dymo_devs[0]
            print(f"Found one Dymo device: {device_info(dev)}")
        dev = dymo_devs[0]
        if dev.idProduct in SUPPORTED_PRODUCTS:
            print(f"Recognized device as {SUPPORTED_PRODUCTS[dev.idProduct]}")
        else:
            print(f"Unrecognized device: {hex(dev.idProduct)}. {UNCONFIRMED_MESSAGE}")

        try:
            dev.get_active_configuration()
            print("Active device configuration already found.")
        except usb.core.USBError:
            try:
                dev.set_configuration()
                print("Device configuration set.")
            except usb.core.USBError as e:
                if e.errno == 13:
                    raise RuntimeError("Access denied")
                if e.errno == 16:
                    print("Device is busy, but this is okay.")
                else:
                    raise

        intf = usb.util.find_descriptor(
            dev.get_active_configuration(), bInterfaceClass=PRINTER_INTERFACE_CLASS
        )
        if intf is not None:
            print(f"Opened printer interface: {repr(intf)}")
        else:
            intf = usb.util.find_descriptor(
                dev.get_active_configuration(), bInterfaceClass=HID_INTERFACE_CLASS
            )
            if intf is not None:
                print(f"Opened HID interface: {repr(intf)}")
            else:
                die("Could not open a valid interface.")
        assert isinstance(intf, usb.core.Interface)

        try:
            if dev.is_kernel_driver_active(intf.bInterfaceNumber):
                print(f"Detaching kernel driver from interface {intf.bInterfaceNumber}")
                dev.detach_kernel_driver(intf.bInterfaceNumber)
        except NotImplementedError:
            print(f"Kernel driver detaching not necessary on " f"{platform.system()}.")
        devout = usb.util.find_descriptor(
            intf,
            custom_match=(
                lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
                == usb.util.ENDPOINT_OUT
            ),
        )
        devin = usb.util.find_descriptor(
            intf,
            custom_match=(
                lambda e: usb.util.endpoint_direction(e.bEndpointAddress)
                == usb.util.ENDPOINT_IN
            ),
        )

        if not devout or not devin:
            die("The device endpoints not be found.")

        # create dymo labeler object
        lm = DymoLabeler(devout, devin, synwait=syn_wait, tape_size=tape_size)

        print("Printing label..")
        lm.printLabel(label_matrix, margin=margin)
        print("Done printing.")
        usb.util.dispose_resources(dev)
        print("Cleaned up.")
